#include "stdafx.h"
#include "windows.h"
#include "avisynth.h"


class DePat : public GenericVideoFilter {
  // DePat defines the name of your filter class.
  // This name is only used internally, and does not affect the name of your filter or similar.
  // This filter extends GenericVideoFilter, which incorporates basic functionality.
  // All functions present in the filter must also be present here.


  PClip WindowVideo; // define the secondary video clip parameter variable
  int SquareSize; // define thesquare size parameter variable


public:
  // This defines that these functions are present in your class.
  // These functions must be that same as those actually implemented.
  // Since the functions are "public" they are accessible to other classes.
  // Otherwise they can only be called from functions within the class itself.

  DePat(PClip _child,      PClip _WindowVideo, int _SquareSize,  IScriptEnvironment* env);
  // This is the constructor. It does not return any value, and is always used,
  //  when an instance of the class is created.
  // Since there is no code in this, this is the definition.

  ~DePat();
  // The is the destructor definition. This is called when the filter is destroyed.


  PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
  // This is the function that AviSynth calls to get a given frame.
  // So when this functions gets called, the filter is supposed to return frame n.
};

class DePatDebug : public GenericVideoFilter {
  // DePatDebug defines the name of your filter class.
  // This name is only used internally, and does not affect the name of your filter or similar.
  // This filter extends GenericVideoFilter, which incorporates basic functionality.
  // All functions present in the filter must also be present here.


  PClip WindowVideo; // define the secondary video clip parameter variable

public:
  // This defines that these functions are present in your class.
  // These functions must be that same as those actually implemented.
  // Since the functions are "public" they are accessible to other classes.
  // Otherwise they can only be called from functions within the class itself.

  DePatDebug(PClip _child, PClip _WindowVideo, IScriptEnvironment* env);
  // This is the constructor. It does not return any value, and is always used,
  //  when an instance of the class is created.
  // Since there is no code in this, this is the definition.

  ~DePatDebug();
  // The is the destructor definition. This is called when the filter is destroyed.


  PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
  // This is the function that AviSynth calls to get a given frame.
  // So when this functions gets called, the filter is supposed to return frame n.
};

class DePatForceYV12toRGB24 : public GenericVideoFilter {
public:

  DePatForceYV12toRGB24(PClip _child, IScriptEnvironment* env);

  ~DePatForceYV12toRGB24();

  PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
  const VideoInfo& __stdcall DePatForceYV12toRGB24::GetVideoInfo();

};












/***************************
 * The following is the implementation
 * of the defined functions.
 ***************************/

//Here is the acutal constructor code used
DePat::DePat(PClip _child, PClip _WindowVideo, int _SquareSize,  IScriptEnvironment* env) :
GenericVideoFilter(_child), WindowVideo(_WindowVideo), SquareSize(_SquareSize){
  // This is the implementation of the constructor.
  // The child clip (source clip) is inherited by the GenericVideoFilter,
  //  where the following variables gets defined:
  //   PClip child;   // Contains the source clip.
  //   VideoInfo vi;  // Contains videoinfo on the source clip.
}

// This is where any actual destructor code used goes
DePat::~DePat() {
  // This is where you can deallocate any memory you might have used.
}




PVideoFrame __stdcall DePat::GetFrame(int n, IScriptEnvironment* env) {
  // This is the implementation of the GetFrame function.
  // See the header definition for further info.

  PVideoFrame src = child->GetFrame(n, env);
  // Request frame 'n' from the child (source) clip.
  PVideoFrame window = WindowVideo->GetFrame(n, env);
  // Request frame "'n" from the WindowVideo clip
  PVideoFrame dst = env->NewVideoFrame(vi);
  // Construct a frame based on the information of the current frame
  // contained in the "vi" struct.

  const unsigned char* srcp = src->GetReadPtr();
  // Request a Read pointer from the source frame.
  // This will return the position of the upperleft pixel in YUY2 images,
  // and return the lower-left pixel in RGB.
  // RGB images are stored upside-down in memory.
  // You should still process images from line 0 to height.

  unsigned char* dstp = dst->GetWritePtr();
  // Request a Write pointer from the newly created destination image.
  // You can request a writepointer to images that have just been
  // created by NewVideoFrame. If you recieve a frame from PClip->GetFrame(...)
  // you must call env->MakeWritable(&frame) be recieve a valid write pointer.

  const int dst_pitch = dst->GetPitch();
  // Requests pitch (length of a line) of the destination image.
  // For more information on pitch see: http://www.avisynth.org/index.php?page=WorkingWithImages
  // (short version - pitch is always equal to or greater than width to allow for seriously fast assembly code)

  const int dst_width = dst->GetRowSize();
  // Requests rowsize (number of used bytes in a line.
  // See the link above for more information.

  const int dst_height = dst->GetHeight();
  // Requests the height of the destination image.

  const int src_pitch = src->GetPitch();
  const int src_width = src->GetRowSize();
  const int src_height = src->GetHeight();

  const unsigned char* windowp=window->GetReadPtr();
  const int window_pitch = window->GetPitch();
  const int window_width = window->GetRowSize();
  const int window_height = window->GetHeight();
  // Get info on the Windowed Clip (see src definitions for more information)


  int w, h;

  if (!vi.IsRGB24()) {
    env->ThrowError("MUST BE RGB24");
  }
  // The code just deals with RGB24 colourspace where each pixel is represented by
  // 3 bytes, Blue, Green and Red.
  // Although this colourspace is the easiest to understand, it is very rarely used because
  // a 3 byte sequence (24bits) cannot be processed easily using normal 32 bit registers.

  for (h=0; h < src_height;h++) {
  	for (w = 0; w < src_width; w+=3) {
	    *(dstp + w) = *(srcp + w);            // Blue
	    *(dstp + w + 1) = *(srcp + w + 1);    // Green
	    *( dstp + w + 2) = *(srcp + w + 2);   // Red
  	}
  	srcp = srcp + src_pitch;
	  dstp = dstp + dst_pitch;
  }
  // end copy src to dst

  //Now draw a white square in the middle of the frame
  // Normally you'd do this code within the loop above but here it is in a separate loop for clarity;

  dstp = dst->GetWritePtr();
  dstp = dstp + (dst_height/2 - SquareSize/2)*dst_pitch;
  for (h=0; h < SquareSize;h++) { // only scan 100 lines
    for (w = dst_width/2 - SquareSize*3/2; w < dst_width/2 + SquareSize*3/2; w+=3) {
      *(dstp + w) = 255;          // Set Blue to maximum value.
	    *(dstp + w + 1) = 255;      // and Green.
      *(dstp + w + 2) = 255;      // and Red - therefore the whole pixel is now white.
    }
    dstp = dstp + dst_pitch;
  }

  return dst;
}

//Here is the acutal constructor code used
DePatDebug::DePatDebug(PClip _child, PClip _WindowVideo,  IScriptEnvironment* env) :
GenericVideoFilter(_child), WindowVideo(_WindowVideo) {
  // This is the implementation of the constructor.
  // The child clip (source clip) is inherited by the GenericVideoFilter,
  //  where the following variables gets defined:
  //   PClip child;   // Contains the source clip.
  //   VideoInfo vi;  // Contains videoinfo on the source clip.
}

// This is where any actual destructor code used goes
DePatDebug::~DePatDebug() {
  // This is where you can deallocate any memory you might have used.
}




PVideoFrame __stdcall DePatDebug::GetFrame(int n, IScriptEnvironment* env) {
  // This is the implementation of the GetFrame function.
  // See the header definition for further info.

  PVideoFrame src = child->GetFrame(n, env);
  // Request frame 'n' from the child (source) clip.
  PVideoFrame window = WindowVideo->GetFrame(n, env);
  // Request frame "'n" from the WindowVideo clip
  PVideoFrame dst = env->NewVideoFrame(vi);
  // Construct a frame based on the information of the current frame
  // contained in the "vi" struct.

  const unsigned char* srcp = src->GetReadPtr();
  // Request a Read pointer from the source frame.
  // This will return the position of the upperleft pixel in YUY2 images,
  // and return the lower-left pixel in RGB.
  // RGB images are stored upside-down in memory.
  // You should still process images from line 0 to height.

  unsigned char* dstp = dst->GetWritePtr();
  // Request a Write pointer from the newly created destination image.
  // You can request a writepointer to images that have just been
  // created by NewVideoFrame. If you recieve a frame from PClip->GetFrame(...)
  // you must call env->MakeWritable(&frame) be recieve a valid write pointer.

  const int dst_pitch = dst->GetPitch();
  // Requests pitch (length of a line) of the destination image.
  // For more information on pitch see: http://www.avisynth.org/index.php?page=WorkingWithImages
  // (short version - pitch is always equal to or greater than width to allow for seriously fast assembly code)

  const int dst_width = dst->GetRowSize();
  // Requests rowsize (number of used bytes in a line.
  // See the link above for more information.

  const int dst_height = dst->GetHeight();
  // Requests the height of the destination image.

  const int src_pitch = src->GetPitch();
  const int src_width = src->GetRowSize();
  const int src_height = src->GetHeight();

  const unsigned char* windowp=window->GetReadPtr();
  const int window_pitch = window->GetPitch();
  const int window_width = window->GetRowSize();
  const int window_height = window->GetHeight();
  // Get info on the Windowed Clip (see src definitions for more information)


  int w, h;

  if (!vi.IsRGB24()) {
    env->ThrowError("MUST BE RGB24");
  }
  // The code just deals with RGB24 colourspace where each pixel is represented by
  // 3 bytes, Blue, Green and Red.
  // Although this colourspace is the easiest to understand, it is very rarely used because
  // a 3 byte sequence (24bits) cannot be processed easily using normal 32 bit registers.

  for (h=0; h < src_height;h++) {
  	for (w = 0; w < src_width; w+=3) {
      unsigned char r1, g1, b1;
      unsigned char r2, g2, b2;
      b1 = *(srcp + w);
      g1 = *(srcp + w + 1);
      r1 = *(srcp + w + 2);
      b2 = *(windowp + w);
      g2 = *(windowp + w + 1);
      r2 = *(windowp + w + 2);
	    *(dstp + w) = abs(b1 - b2);
	    *(dstp + w + 1) = abs(g1 - g2);
	    *( dstp + w + 2) = abs(r1 - r2);
  	}
  	srcp = srcp + src_pitch;
	  dstp = dstp + dst_pitch;
	  windowp = windowp + dst_pitch;
  }
  // end copy src to dst

  return dst;
}

//Here is the acutal constructor code used
DePatForceYV12toRGB24::DePatForceYV12toRGB24(PClip _child, IScriptEnvironment* env) :
GenericVideoFilter(_child) {
  // This is the implementation of the constructor.
  // The child clip (source clip) is inherited by the GenericVideoFilter,
  //  where the following variables gets defined:
  //   PClip child;   // Contains the source clip.
  //   VideoInfo vi;  // Contains videoinfo on the source clip.

  if (!vi.IsYV12()) {
    env->ThrowError("MUST BE YV12");
  }

  vi.pixel_type = VideoInfo::CS_BGR24;
}

// This is where any actual destructor code used goes
DePatForceYV12toRGB24::~DePatForceYV12toRGB24() {
  // This is where you can deallocate any memory you might have used.
}

const VideoInfo& __stdcall DePatForceYV12toRGB24::GetVideoInfo() {
  static VideoInfo new_vi = VideoInfo(vi);
  //new_vi.pixel_type = VideoInfo::CS_BGR24;
  return new_vi;
}



PVideoFrame __stdcall DePatForceYV12toRGB24::GetFrame(int n, IScriptEnvironment* env) {
  // This is the implementation of the GetFrame function.
  // See the header definition for further info.

  // PVideoFrame prev_frame = child->GetFrame(n==0 ? 0 : n-1, env);
  //PVideoFrame prev_frame = child->GetFrame(n, env);

  PVideoFrame cur_frame = child->GetFrame(n, env);

  // PVideoFrame next_frame = child->GetFrame(n+1, env);
  //PVideoFrame next_frame = child->GetFrame(n, env);

  //const unsigned char* prev_p = prev_frame->GetReadPtr();
  const unsigned char* cur_p = cur_frame->GetReadPtr();
  //const unsigned char* next_p = next_frame->GetReadPtr();

  const int width = vi.width;
  const int height = vi.height;
  const int offsetU = cur_frame->GetOffset(PLANAR_U);
  const int offsetV = cur_frame->GetOffset(PLANAR_V);
  const int src_row_size = cur_frame->GetRowSize(PLANAR_Y);
  const int src_row_sizeUV = cur_frame->GetRowSize(PLANAR_U);
  const int src_pitch = cur_frame->GetPitch(PLANAR_Y);
	const int src_pitchUV = cur_frame->GetPitch(PLANAR_U);
	const int src_heightUV = cur_frame->GetHeight(PLANAR_U);

  PVideoFrame dst = env->NewVideoFrame(vi);
  env->MakeWritable(&dst);

  const int dst_row_size = dst->GetRowSize();
  const int dst_pitch = dst->GetPitch();
  unsigned char* dstp = dst->GetWritePtr();

  int x, y;
  for (y=0; y < height;y++) {
  	for (x = 0; x < width; x++) {
      unsigned char color_y, color_u, color_v;
      unsigned char color_r, color_g, color_b;
      const unsigned char *src_p;
      unsigned char *dst_p;
      int xY,yY;
      int xU,yU;
      int xV,yV;
      xY = x;
      yY = height-y-1;
      xU = x / 2;
      yU = yY / 2;
      xV = x / 2;
      yV = yY / 2;

      src_p = cur_p;
      color_y = *(src_p +           (yY*src_pitch)   + xY);
      color_u = *(src_p + offsetU + (yU*src_pitchUV) + xU);
      color_v = *(src_p + offsetV + (yV*src_pitchUV) + xV);

      //
      color_r = color_v;
      color_g = color_y;
      color_b = color_u;

      //
      dst_p = dstp;
	    *(dst_p + (y*dst_row_size) + (x * 3)) = color_b;
	    *(dst_p + (y*dst_row_size) + (x * 3)+1) = color_g;
	    *(dst_p + (y*dst_row_size) + (x * 3)+2) = color_r;
  	}
  }

  return dst;
}

// This is the function that created the filter, when the filter has been called.
// This can be used for simple parameter checking, so it is possible to create different filters,
// based on the arguments recieved.

AVSValue __cdecl Create_DePat(AVSValue args, void* user_data, IScriptEnvironment* env) {
  return new DePat(args[0].AsClip(), // the 0th parameter is the source clip
                          args[1].AsClip(), // Corresponds to our 1st parameter - the WindowedVideo
                          args[2].AsInt(0), // Corresponds to our 2nd parameter - the size of the square in pixels.
                          env);
  // Calls the constructor with the arguments provied.
}

AVSValue __cdecl Create_DePatDebug(AVSValue args, void* user_data, IScriptEnvironment* env) {
  return new DePatDebug(args[0].AsClip(),
                          args[1].AsClip(),
                          env);
  // Calls the constructor with the arguments provied.
}

AVSValue __cdecl Create_DePatForceYV12toRGB24(AVSValue args, void* user_data, IScriptEnvironment* env) {
  return new DePatForceYV12toRGB24(args[0].AsClip(),
                          env);
  // Calls the constructor with the arguments provied.
}

// The following function is the function that actually registers the filter in AviSynth
// It is called automatically, when the plugin is loaded to see which functions this filter contains.

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) {
  env->AddFunction("DePat", "cc[SIZE]i", Create_DePat, 0);
  env->AddFunction("DePatDebug", "cc", Create_DePatDebug, 0);
  env->AddFunction("DePatForceYV12toRGB24", "c", Create_DePatForceYV12toRGB24, 0);

  return "`DePat' DePat plugin";
}
